from infi.pyutils.lazy import cached_function, clear_cache
from jira import JIRAError
from munch import Munch
from logging import getLogger
from functools import partial
from .config import Configuration
import requests
from infi.pyutils.lazy import cached_function
from .config import Configuration
from .credential_store import JIRACredentialsStore
from requests.auth import HTTPBasicAuth


logger = getLogger(__name__)


CURRENT_USER = "currentUser()"
ASSIGNED_ISSUES = "{}assignee = {} AND resolution = unresolved ORDER BY priority DESC, created ASC"


@cached_function
def get_auth(fqdn):
    config = Configuration.from_file()
    credential_store = JIRACredentialsStore()
    credentials = credential_store.get_credentials(fqdn)
    return HTTPBasicAuth(credentials.get_username(), credentials.get_password())


@cached_function
def get_headers():
    return {'Accept': 'application/json'}


@cached_function
def get_jira():
    from jira import JIRA as _JIRA

    class JIRA(_JIRA):
        def __del__(self):
            # workaround for silencing the exception thrown upon exit to stderr
            # Exception AttributeError: "'NoneType' object has no attribute 'version_info'" in <bound method JIRA.__del__ of <jira.client.JIRA object at 0x10e46a250>> ignored
            pass

    config = Configuration.from_file()
    options = dict(server="https://{0}".format(config.jira_fqdn))
    basic_auth = get_auth(config.jira_fqdn)
    return JIRA(options, basic_auth=(basic_auth.username, basic_auth.password))


@cached_function
def get_custom_fields():
    return {item['name']: item['id'] for item in get_jira().fields() if item['custom']}


@cached_function
def get_custom_fields_schema():
    return {item['name']: item['schema']['custom'] for item in get_jira().fields() if item['custom']}


@cached_function
def get_issues__assigned_to_user(user, project=None):
    return get_jira().search_issues(ASSIGNED_ISSUES.format("project={} AND ".format(project) if project else '', user))


@cached_function
def get_issues__assigned_to_me(project=None):
    return get_issues__assigned_to_user(CURRENT_USER, project)


def add_labels_to_issue(key, labels):
    issue = get_issue(key)
    labels = set.union(set([str(label) for label in labels]), set(issue.fields().labels))
    issue.update(labels=[dict(add=label) for label in labels])


def assign_issue(key, assignee):
    get_jira().assign_issue(key, assignee)


def from_jira_formatted_datetime(formatted_string):
    # http://stackoverflow.com/questions/127803/how-to-parse-iso-formatted-date-in-python
    import re
    import datetime
    return datetime.datetime(*list(map(int, re.split('[^\d]', formatted_string)[:-1])))


def from_jira_formatted_date(formatted_string):
    return from_jira_formatted_datetime(formatted_string+'T00:00:00.000+0000')


def to_jira_formatted_date(datetime_object):
    return datetime_object.strftime("%Y-%m-%d")


def matches(str_a, str_b):
    return str_a is not None and str_b is not None and str_a.lower() == str_b.lower()


def transition_issue(key, transition_string, additional_fields, id_lookup_method=None, **kwargs):
    jira = get_jira()
    issue = jira.issue(key)
    issue_type_name = issue_mappings.Type(issue)
    project_key = issue.fields().project.key
    [transition] = [item['id'] for item in jira.transitions(issue) if matches(item['name'], transition_string)]
    fields = dict()
    if additional_fields:
        for key, value in list(additional_fields.items()):
            if key in ('issuelinks', ):
                fields[key] = value
            else:
                fields[get_custom_fields()[key]] = _compute_value(key, value, id_lookup_method)
    logger.debug("calling transition_issue(issue={issue!r}, transition={transition!r}, fields={fields!r})".format(issue=issue, transition=transition, fields=fields))
    jira.transition_issue(issue=issue.key, transition=transition, fields=fields, **kwargs)


def resolve_issue(key, resolution_string, fix_versions_strings):
    jira = get_jira()
    issue = jira.issue(key)
    [resolution] = [item.id for item in jira.resolutions() if matches(item.name, resolution_string)]
    project_versions = jira.project_versions(issue.fields().project)
    fix_versions = [dict(id=item.id) for item in project_versions if item.name in fix_versions_strings]
    fields = dict(resolution=dict(id=resolution), fixVersions=fix_versions)
    transition_issue(key, "Resolve Issue", dict(), **fields)


def start_progress(key):
    transition_issue(key, "Start Progress", dict())


def stop_progress(key):
    transition_issue(key, "Stop Progress", dict())


def reopen(key):
    transition_issue(key, "Reopen Issue", dict())


def iter_projects():
    for item in get_jira().projects():
      yield get_project(item.key)


@cached_function
def get_project(key):
    return get_jira().project(key.upper())


@cached_function
def get_version(key, name):
    [version] = [version for version in get_project(key).versions if version.name == name]
    return version


@cached_function
def get_next_release_name_for_issue(key):
    jira = get_jira()
    issue = jira.issue(key)
    project = issue.fields().project.key
    return get_next_release_name_in_project(project)


@cached_function
def get_next_release_name_in_project(key):
    project = get_project(key)
    next_releases = sorted([version for version in project.versions if not version.released],
                           key=lambda version: from_jira_formatted_date(getattr(version, "releaseDate", '2121-12-12')))
    if next_releases:
      return next_releases[0].name
    return ''


@cached_function
def _get_options(customfield_name):
    from .custom_field_editor import get_options_for_custom_field
    config = Configuration.from_file()
    customfield_id = get_custom_fields()[customfield_name]
    return get_options_for_custom_field(customfield_id)


def get_custom_field_values(customfield_name):
    options = _get_options(customfield_name)
    return [item['optionvalue'] for item in options]


def get_enabled_custom_field_values(customfield_name):
    options = _get_options(customfield_name)
    return [item['optionvalue'] for item in options if not item['disabled']]


def get_custom_field_value_id_from_createmeta(project_key, issue_type_name, key, value):
    result = get_jira().createmeta(issuetypeNames=[issue_type_name], projectKeys=[project_key], expand=['projects.issuetypes.fields'])
    values = result['projects'][0]['issuetypes'][0]['fields'][get_custom_fields()[key]]['allowedValues']
    [value_id] = [item['id'] for item in values if item['value'] == value]
    return value_id


def _compute_value(key, value, id_lookup_method):
    def _translate(value):
        if key in ('issuelinks', ):
          return value
        if 'select' in get_custom_fields_schema()[key]:
          return {'value': value}
        if 'radiobuttons' in get_custom_fields_schema()[key]:
          return {'id': id_lookup_method(key, value)}
        if 'multicheckboxes' in get_custom_fields_schema()[key]:
          return {'id': id_lookup_method(key, value)}
        if 'userpicker' in get_custom_fields_schema()[key]:
          return {'name': value}
        return value

    if isinstance(value, (list, tuple)):
      return [_translate(item) for item in value]
    result = _translate(value)
    return result if isinstance(result, dict) else result


def create_issue(project_key, issue_type_name, component_name, fix_version_name, details, priority=None, assignee=None, parent=None, additional_fields=None, id_lookup_method=None):
    jira = get_jira()
    project = jira.project(project_key)
    [issue_type] = [issue_type for issue_type in project.issueTypes
                                    if matches(issue_type.name, issue_type_name)]
    components = [component for component in project.components
                                if matches(component.name, component_name)]
    versions = [version for version in project.versions
                            if matches(version.name, fix_version_name)]
    summary = details.split("\n", 1)[0]
    description = details.split("\n", 1)[1:]
    fields = dict(project=dict(id=str(project.id)),
                  issuetype=dict(id=str(issue_type.id)),
                  components=[dict(id=str(component.id)) for component in components],
                  fixVersions=[dict(id=str(version.id)) for version in versions],
                  summary=summary, description=description[0] if description else None)
    if not fields['description']:
      fields.pop('description')
    if assignee:
      fields['assignee'] = dict(name=assignee)
    if priority:
      fields['priority'] = dict(name=priority)
    if parent:
      fields['parent'] = dict(key=parent)
    if not versions:
        fields.pop('fixVersions')
    if not components:
        fields.pop('components')
    if additional_fields:
        for key, value in list(additional_fields.items()):
            _id_lookup_method = id_lookup_method or partial(get_custom_field_value_id_from_createmeta, project_key=project_key, issue_type_name=issue_type_name)
            fields[get_custom_fields()[key]] = _compute_value(key, value, _id_lookup_method)
    issue = jira.create_issue(fields=fields)
    return issue


def create_link(link_type_name, from_key, to_key):
    jira = get_jira()
    [link_type] = [link_type for link_type in jira.issue_link_types()
                                 if matches(link_type.name, link_type_name)]

    kwargs = dict(type=link_type.name, inwardIssue=from_key, outwardIssue=to_key)
    jira.create_issue_link(**kwargs)


def search_issues(query, maxResults=2000, expand=None):
    jira = get_jira()
    return jira.search_issues(query, maxResults=maxResults, expand=expand)


def comment_on_issue(key, message):
    get_jira().add_comment(issue=key, body=message)


@cached_function
def get_issue(key):
    return get_jira().issue(key.upper())


@cached_function
def get_query_by_filter(name):
    """:returns: the jql of the filter"""
    for _filter in get_jira().favourite_filters():
        if _filter.name == name:
            return _filter.jql
    raise JIRAError(404, "no such filter")


issue_mappings = Munch(Rank=lambda issue: int(issue.fields().customfield_10700),
                       Type=lambda issue: issue.fields().issuetype.name,
                       Key=lambda issue: issue.key,
                       Summary=lambda issue: issue.fields().summary,
                       Description=lambda issue: issue.fields().description,
                       Priority=lambda issue: issue.fields().priority.name,
                       Project=lambda issue: issue.fields().project.name,
                       Status=lambda issue: issue.fields().status.name,
                       Resolution=lambda issue: (issue.fields().resolution or Munch(name="Unresolved")).name,
                       Created=lambda issue: from_jira_formatted_datetime(issue.fields().created),
                       Updated=lambda issue: from_jira_formatted_datetime(issue.fields().updated),
                       Assignee=lambda issue: issue.fields().assignee.displayName,
                       Reporter=lambda issue: issue.fields().reporter.displayName,
                       Labels=lambda issue: issue.fields().labels,
                       Comments=lambda issue: issue.fields().comment.comments,
                       AffectsVersions=lambda issue: [item.name for item in getattr(issue.fields(), 'versions', list())],
                       FixVersions=lambda issue: [item.name for item in getattr(issue.fields(), 'fixVersions', list())],
                       Components=lambda issue: [item.name for item in issue.fields().components],
                       IssueLinks=lambda issue: issue.fields().issuelinks,
                       SubTasks=lambda issue: issue.fields().subtasks,
                       Attachments=lambda issue: issue.fields().attachment,
                       )


@cached_function
def is_user_exists(username):
    return any(user.key == username for user in get_jira().search_users(username))


@cached_function
def get_user_by_name(name):
     users = get_jira().search_users(name)
     return users[0].key if len(users) == 1 else None


from .custom_field_editor import get_options_for_custom_field, get_custom_field_id_by_name



"""jirelnotes
infinidat jira/confluence release-notes command-line tool

Usage:
    jirelnotes show {project} [--include-next-release]
    jirelnotes publish {project} [--include-next-release]
    jirelnotes notify {project} {version} [<other-versions-included-in-this-releases>...] [--dry-run]
    jirelnotes fetch {project}
    jirelnotes config set <confluence_fqdn>

Options:
    show                                 show the generated release notes
    publish                              publish the release notes to the wiki
    fetch                                print the release note from the wiki
    notify                               find linked issues and put comments
    config set                           set the Confluence IP or hostname. The credentials are taken from jissue config
    --project=PROJECT                    project key {project_default}
"""
from __future__ import print_function

RELEASE_NOTES_TITLE_KEY = 'Release Notes Title'
RELEASE_NOTES_DESCRIPTION_KEY = 'Release Notes Description'
NOTIFICATION_MESSAGE = """{{ project.name }} v{{ version }} has been released and now available for download.
{% if other_versions %}
This release supersedes the following unreleased versions:
{%- for version in other_versions %}
* {{ version }}
{%- endfor %}
{%- endif %}
{% if resolved_issues %}
This release solves the following related issues:
{%- for issue in resolved_issues %}
* {{ issue.key }}: {{ issue_mappings.Summary(issue) }}
{%- endfor %}
{%- endif %}
{% if unresolved_issues %}
Still, there are more related issues that are still open:
{%- for issue in unresolved_issues %}
* {{ issue.key }}: {{ issue_mappings.Summary(issue) }}
{%- endfor %}
{%- else %}
All related issues are now resolved.
{%- endif %}
"""


def _get_arguments(argv, environ):
    from .__version__ import __version__
    from docopt import docopt
    from munch import Munch
    project_default = "[default: {}]".format(environ["JISSUE_PROJECT"]) if "JISSUE_PROJECT" in environ else ""
    version_default = "[default: {}]".format(environ["JISSUE_VERSION"]) if "JISSUE_VERSION" in environ else ""
    doc_with_defaults = __doc__.format(project_default=project_default, version_default=version_default,
                                       project="[--project=PROJECT]" if project_default else "--project=PROJECT",
                                       version="[--release=RELEASE]" if version_default else "--release=RELEASE")
    arguments = Munch(docopt(doc_with_defaults, argv=argv, help=True, version=__version__))
    if environ.get("JISSUE_PROJECT") and not arguments.get("--project"):
        arguments["--project"] = environ["JISSUE_PROJECT"]
    if environ.get("JISSUE_VERSION") and not arguments.get("--release"):
        arguments["--release"] = environ["JISSUE_VERSION"]
    return arguments


def get_field(issue, key):
    from .jira_adapter import get_custom_fields
    result = getattr(issue.fields(), get_custom_fields()[key])
    return result.replace('\r', '') if result else None


def get_issue_details(issue):
    return dict(key=issue.key,
                title=get_field(issue, RELEASE_NOTES_TITLE_KEY),
                description=get_field(issue, RELEASE_NOTES_DESCRIPTION_KEY))


def is_bug(issue):
    from .jira_adapter import issue_mappings
    return issue_mappings['Type'](issue) == "Bug"


def is_improvement(issue):
    from .jira_adapter import issue_mappings
    return issue_mappings['Type'](issue) == "Improvement"


def is_new_feature(issue):
    from .jira_adapter import issue_mappings
    return issue_mappings['Type'](issue) == "New Feature"


def get_release_notes_contents_for_specfic_version(project, version):
    from .jira_adapter import get_custom_fields, get_jira
    release_date = getattr(version, 'releaseDate', '')
    base_query = "project={} AND fixVersion={!r} AND {!r} IS NOT EMPTY".format(project.key, str(version.name), RELEASE_NOTES_TITLE_KEY)
    known_issues_query = "project={0} AND {1!r} IS NOT EMPTY AND (" \
                         "(fixVersion IN ('known issues') AND status IN (Open, Reopened)) OR " \
                         "(affectedVersion in ({2!r})) OR " \
                         "(labels=known-issue AND status WAS IN (Resolved) ON {3} AND fixVersion > {2!r}))"
    known_issues_query = known_issues_query.format(project.key, RELEASE_NOTES_TITLE_KEY,
                                                   str(version.name), release_date or 'now()')
    resolved_issues = list(get_jira().search_issues(base_query))
    known_issues = list(get_jira().search_issues(known_issues_query))
    if resolved_issues or known_issues_query:
        topics = [dict(name="What's new in this release", issues=[get_issue_details(issue) for issue in resolved_issues if is_new_feature(issue)]),
                  dict(name='Improvements', issues=[get_issue_details(issue) for issue in resolved_issues if is_improvement(issue)]),
                  dict(name='Fixed issues', issues=[get_issue_details(issue) for issue in resolved_issues if is_bug(issue)]),
                  dict(name='Known issues', issues=[get_issue_details(issue) for issue in known_issues]),
                  ]
        return dict(name=version.name, release_date=release_date, topics=topics)
    else:
        return dict()


def should_appear_in_release_notes(release):
    return bool(release) and any(topic['issues'] for topic in release['topics'])



def render_release_notes(project_key, project_version, include_next_release, page_id, header_id, footer_id):
    from re import match
    from jinja2 import Template
    from pkg_resources import resource_string
    from .jira_adapter import get_project, get_version, get_jira, issue_mappings, get_custom_fields
    from .jira_adapter import get_next_release_name_in_project
    from .confluence_adapter import iter_attachments, get_page_contents, get_page_storage
    template = Template(resource_string('infi.jira_cli', 'release_notes.html'))
    project = get_project(project_key)
    real_versions = [version for version in reversed(project.versions) if
                     match(r'[0-9\.]+', version.name) and not version.archived and
                     version.released or (version.name == get_next_release_name_in_project(project_key) if include_next_release else False)]
    releases = [get_release_notes_contents_for_specfic_version(project, version) for version in real_versions]
    attachments = list(iter_attachments(page_id))
    exposed_releases = [release for release in releases if should_appear_in_release_notes(release)]
    return template.render(project=project, releases=exposed_releases, page_id=page_id, attachments=attachments,
                           header=get_page_storage(header_id) if header_id else None,
                           footer=get_page_storage(footer_id) if footer_id else None)


def get_release_notes(project_key, project_version, include_next_release):
    from .confluence_adapter import get_release_notes_page_id
    from .confluence_adapter import get_release_notes_header_page_id, get_release_notes_footer_page_id
    page_id = get_release_notes_page_id(project_key)
    header_id = get_release_notes_header_page_id(project_key)
    footer_id = get_release_notes_footer_page_id(project_key)
    release_notes = render_release_notes(project_key,
                                         project_version,
                                         include_next_release,
                                         page_id,
                                         header_id,
                                         footer_id)
    return release_notes, page_id


def publish_release_notes(project_key, project_version, include_next_release):
    from .confluence_adapter import update_page_contents
    release_notes, page_id = get_release_notes(project_key, project_version, include_next_release)
    update_page_contents(page_id, release_notes)


def show_release_notes(project_key, project_version, include_next_release):
    release_notes, page_id = get_release_notes(project_key, project_version, include_next_release)
    print(release_notes)


def fetch_release_notes(project_key):
    from .confluence_adapter import get_page_contents, get_release_notes_page_id
    print(get_page_contents(get_release_notes_page_id(project_key)))


def notify_related_tickets(project_key, project_version, other_versions, dry_run):
    def _build_jira_query_string():
        from pkg_resources import parse_version
        if other_versions:
            versions.extend(sorted(set(other_versions + [project_version]), key=lambda version: parse_version(version)))
            fix_version_string = 'fixVersion in ({})'.format(', '.join([repr(version) for version in versions]))
        else:
            fix_version_string = 'fixVersion={!r}'.format(project_version)
        return "project={} AND {} AND resolution=Fixed".format(project_key, fix_version_string)

    def _iter_related_tickets(issue):
        for link in issue_mappings.IssueLinks(issue):
            if not link.type.name in ('Originates', 'Clones', 'Cloners', 'Relates', 'Supersede', 'Duplicate'):
                continue
            related_issue = getattr(link, 'inwardIssue', getattr(link, 'outwardIssue', None))
            if related_issue.key.startswith(project_key.upper()):
                continue
            yield related_issue

    def find_issues_in_other_projects_that_are_pending_on_this_release():
        related_tickets = {}
        for issue in search_issues(_build_jira_query_string()):
            for related_ticket in _iter_related_tickets(issue):
                related_tickets.setdefault(get_issue(related_ticket.key), list()).append(issue)
        return related_tickets

    def _iter_related_remaining_open_issues(related_ticket):
        for link in issue_mappings.IssueLinks(related_ticket):
            if not hasattr(link, 'outwardIssue'):
                continue
            if not issue_mappings.Status(link.outwardIssue) in ('Open', 'Reopened'):
                continue
            yield link.outwardIssue

    def _build_comment(resolved_issues, unresolved_issues):
        from jinja2 import Template
        sort_issues = lambda issues: sorted(issues, key=lambda issue: issue.key)
        notification_template = Template(NOTIFICATION_MESSAGE.strip())
        return notification_template.render(project=project,
                                            version=project_version,
                                            other_versions=versions[:-1],
                                            resolved_issues=sort_issues(resolved_issues),
                                            unresolved_issues=sort_issues(unresolved_issues),
                                            issue_mappings=issue_mappings)

    from .jira_adapter import search_issues, issue_mappings, comment_on_issue, get_project, get_issue
    project = get_project(project_key)
    versions = []
    related_tickets = find_issues_in_other_projects_that_are_pending_on_this_release()
    for related_ticket, resolved_issues in sorted(list(related_tickets.items()), key=lambda item: item[0].key):
        unresolved_issues = list(_iter_related_remaining_open_issues(related_ticket))
        comment = _build_comment(resolved_issues, unresolved_issues)
        comment = "".join(i for i in comment if ord(i)<128)
        if dry_run:
            print("<--- COMMENT ON {0} STARTS HERE --->\n{1}\n<--- COMMENT ON {0} ENDS HERE ----->".format(related_ticket.key, comment))
        else:
            print('commenting on %s' % related_ticket.key)
            comment_on_issue(related_ticket.key, comment)


def config_set(confluence_fqdn):
    from .config import Configuration
    config = Configuration.from_file()
    config.confluence_fqdn = confluence_fqdn
    config.save()


def do_work(arguments):
    project_key = arguments['--project']
    project_version = arguments.get('--release')
    if arguments['show']:
        show_release_notes(project_key, project_version, arguments['--include-next-release'])
    elif arguments['publish']:
        publish_release_notes(project_key, project_version, arguments['--include-next-release'])
    elif arguments['fetch']:
        fetch_release_notes(project_key)
    elif arguments['notify']:
        other_versions = arguments.get('<other-versions-included-in-this-releases>', list())
        notify_related_tickets(project_key, project_version, other_versions, arguments['--dry-run'])
    elif arguments['config'] and arguments['set']:
        config_set(arguments['<confluence_fqdn>'])


def _jiject(argv, environ):
    from sys import stderr
    from copy import deepcopy
    from jira import JIRAError
    from infi.execute import ExecutionError
    from .actions import choose_action
    from docopt import DocoptExit
    try:
        arguments = _get_arguments(argv, dict(deepcopy(environ)))
        return do_work(arguments)
    except DocoptExit as e:
        print(e, file=stderr)
        return 1
    except AssertionError as e:
        print(e, file=stderr)
        return 1
    except SystemExit as e:
        print(e, file=stderr)
        return 0
    except JIRAError as e:
        print(e, file=stderr)
    except ExecutionError as e:
        print(e.result.get_stderr(), file=stderr)
    return 1


def main():
    from os import environ
    from sys import argv
    return _jiject(argv[1:], environ)
